//------------------------------------------------------------------------------
//
//  npnode.c
//
// ANTZ - 3D data viewer with realtime real-world interaction based on NPE.
//
// ANTZ is hosted at http://temporalzone.com and NPE at http://neuralphysics.org
//
// Written in 2010-2011 by Shane Saxon - contact@saxondigital.net
//
// Additional code contributed by: Mark Buchanan and Lucas Erickson.
//
// To the extent possible under law, the author(s) have dedicated all copyright 
// and related and neighboring rights to this software to the public domain
// worldwide. This software is distributed without any warranty.
//
// Released under the CC0 license, which is GPL compatible.
//
// You should have received a copy of the CC0 Public Domain Dedication along
// with this software (license file named COPYING.txt). If not, see
// http://creativecommons.org/publicdomain/zero/1.0/
//
//------------------------------------------------------------------------------

#include "npnode.h"

#include "../npctrl.h"


void InitNodeSurface (NPnodePtr node);
void InitNodePoints (NPnodePtr node);
void InitNodePin (NPnodePtr node);
void InitNodeVideo (NPnodePtr node);
void InitNodeGrid (NPnodePtr node);


//------------------------------------------------------------------------------
void npInitNode (void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;

	return;
}


//------------------------------------------------------------------------------
void npCloseNode(void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;

	for (i=0; i < data->map.nodeRootCount; i++)
		npNodeDelete (data->map.node[i], dataRef);
}


// should make this static and add locking for delete to be thread safe,      debug zz
// creates a new node and attaches it as a child of the nodeParent
// if nodeParent is NULL then creates a root node
//------------------------------------------------------------------------------
void*  npNodeNew (int nodeType, NPnodePtr nodeParent, void* dataRef)
{
	int nodeRootIndex = 0;
	NPnodePtr node = NULL;

	pData data = (pData) dataRef;

	// check to see if max total nodes hit
	if (data->map.nodeCount >= kNPnodeMax)
	{
		printf ("err 4487 - kNPnodeMax hit, max new nodes hit limit\n");
		return NULL;
	}

	if (nodeParent != NULL)
	{
		if (nodeParent->childCount >= kNPnodeChildMax)
		{
			printf ("err 4488 - kNPnodeChildMax hit, cannot create child node\n");
			return NULL;
		}
	}
	else
	{
		if (data->map.nodeRootCount >= kNPnodeRootMax)
		{
			printf ("err 4489 - kNPnodeRootMax - max root nodes hit limit\n");
			return NULL;
		}
	}

	//allocate memory to create node and init defaults
	node = (NPnodePtr) malloc (sizeof(NPnode));
	if (data == NULL)
	{
		printf ("err 4490 - malloc failed, cannot create node\n");
		return NULL;
	}

	npInitNodeDefault (node);
			
	// if root node then insert into node array and select it
	if (nodeParent == NULL)
	{
		//insert into the node root array
		data->map.node[data->map.nodeRootCount] = node;

		//select the new node and make it active
		data->map.nodeRootIndex = data->map.nodeRootCount;
		data->map.selectedIndex = data->map.nodeRootIndex;
		data->map.selectedNode = data->map.node[data->map.nodeRootIndex];
		data->map.activeNode = data->map.selectedNode;
		
		data->map.nodeRootCount++;
	}
	else	// attach child to nodeParent
	{

		node->parent = nodeParent;
		node->branchLevel = nodeParent->branchLevel + 1;

		if (nodeParent != data->map.node[kNPnodeRootNull]) //debug, zz
		{
			nodeParent->child[nodeParent->childCount] = node;
			nodeParent->childIndex = nodeParent->childCount;
			nodeParent->childCount++;
		}
	}

	data->map.nodeCount++;

	//set the unique properties, give it an ID and type specific params
	node->id = npNewNodeID();
	node->type = nodeType;

	switch (nodeType)
	{
		case kNodeCamera :	
			npInitNodeCamera (node);
			data->map.activeCam = node; 
			break;
		case kNodeGrid :	InitNodeGrid (node); break;
		case kNodeVideo :	InitNodeVideo (node); break;
		case kNodeSurface :	InitNodeSurface (node); break;
		case kNodePoints :	InitNodePoints (node); break;
		case kNodePin :		
			InitNodePin (node);
			if (node->branchLevel == 0)		//auto offset position
				node->translate.x += 5.0f * data->map.nodeRootCount -35.0f;
			else
			{
				node->geometry = kNPglutSolidTorus;		// workaround?, debug, zz
		
				if (node->branchLevel == 1)
					node->translate.z += -0.5f * (float)(nodeParent->childCount - 1);
				else
				{		// distribute children at 15 deg spacing... is 0.2618 radians
					node->rotateRad.x = 0.2617994f * (float)(nodeParent->childCount - 1);
					if (nodeParent->childCount > 24)
						node->rotateRad.x += 0.1308997f;
				}
			}
			break;
		default : 
			break;
	}

	//creates a map that maps node ID back to a ptr to the node, debug zz
	if (node->id < kNPnodeMax)
		data->map.nodeID[node->id] = node;
	else
		printf("err 4289 - node ID exceeds size of kNPnodeMax\n");


	return (void*)node;
}


//------------------------------------------------------------------------------
void npNodeDelete (NPnodePtr node, void* dataRef)
{
	npNodeRemove (1, node, dataRef);
}

// removes node from tree and recursively traverses tree to free memory
//------------------------------------------------------------------------------
void npNodeRemove (int freeNode, NPnodePtr node, void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;
	NPnodePtr parent = NULL;

	if (node == NULL)
		return;


	//update data structure
	if (node->branchLevel > 0)	//process child node
	{
		parent = node->parent;
		
		if (node->parent != NULL)		//if not orphan node
		{
			//remove this, update to not require child nodes			debug, zz
			if ( node->type == kNodeCamera && parent->childCount < 2 )
				return;

			//compact the gap, update child info
			parent->childCount--;
			for (i = parent->childIndex; i < parent->childCount; i++)
				parent->child[i] = parent->child[i+1];

			parent->child[parent->childCount] = NULL;

			if (parent->childIndex >= parent->childCount
				&& parent->childCount > 0 )
				parent->childIndex = parent->childCount - 1;
		}
	}
	else	//process root node
	{
		//do not remove root null, camera or grid
		if (   node == data->map.node[kNPnodeRootNull] 
			|| node == data->map.node[kNPnodeRootCamera]
			|| node == data->map.node[kNPnodeRootGrid] )
			return;

		//compact the gap by updating root node array, could optimize, zz
		data->map.nodeRootCount--;
		for (i = data->map.nodeRootIndex; i < data->map.nodeRootCount; i++)
			data->map.node[i] = data->map.node[i+1];

		//set the leftover end of array pointer to NULL
		data->map.node[data->map.nodeRootCount] = NULL;

		if (data->map.nodeRootIndex >= data->map.nodeRootCount)
			data->map.nodeRootIndex = data->map.nodeRootCount - 1;
	}

	//free all node data including child nodes on the tree
	if (freeNode)
	{
		// recursive call to free memory all sub-child nodes in tree
		for (i=0; i < node->childCount; i++)
			npNodeRemove(freeNode, node->child[i], dataRef);

		data->map.nodeCount--;

		// now free the memory
		free (node->data);
		free (node);
	}
	
	//restore active and selection index, after recursive call
	if (node->branchLevel == 0)
	{
		data->map.selectedIndex = data->map.nodeRootIndex;
		data->map.selectedNode = data->map.node[data->map.selectedIndex];
		data->map.activeNode = data->map.selectedNode;
	}
	else if (parent != NULL)
	{
		if (parent->childCount <= 0)
		{
			data->map.selectedNode = parent;
			data->map.activeNode = data->map.selectedNode;
		}
		else
		{
			data->map.selectedNode = parent->child[parent->childIndex];
			data->map.activeNode = data->map.selectedNode;
		}
	}

//	printf ("selectIndex: %d   rootCount: %d\n", 
//		data->map.selectedIndex, data->map.nodeRootCount);
}


//------------------------------------------------------------------------------
void npInitNodeDefault (NPnodePtr node)
{
	int i = 0;

//	node->id = npNewNodeID();	// called by npNew and preserved during reset

	node->type			= kNodeDefault;		//node type, point, pin, cam, video

	node->data			= NULL;				//node type specific data

	node->selected		= false;			//defualt is not selected
	
	node->parent		= NULL;				//parent node, binary tree of nodes
	node->branchLevel	= 0;			//0 is the trunk pin (ice-cream cone)

	for (i=0; i < kNPnodeChildMax; i++)
		node->child[i] = NULL;

	node->childIndex	= 0;
	node->childCount	= 0;				//no children by default

	node->channel.x		= kChannelNull;		//channel to use for x,y,z points
	node->channel.y		= kChannelNull;
	node->channel.z		= kChannelNull;
	
	node->channelMapID		= 55;			//maps channels to antz fields, debug zz
	
	node->channelIndex.x = 0;				//last updated channel index
	node->channelIndex.y = 0;
	node->channelIndex.z = 0;

	node->averageType.x	= kAverageNull;		//averaging type applied to data
	node->averageType.y	= kAverageNull;
	node->averageType.z	= kAverageNull;
	
	node->sampleInterval.x = 1;				//the interval to be averaged
	node->sampleInterval.y = 1;
	node->sampleInterval.z = 1;

	node->rotate.angle	= 0.0f;				//orientation vector
	node->rotate.x		= 0.0f;
	node->rotate.y		= 0.0f;
	node->rotate.z		= 1.0f;
	
	node->scale.x		= 1.0f;				//scale node XYZ
	node->scale.y		= 1.0f;
	node->scale.z		= 1.0f;
	
	node->translate.x	= 0.0f;				//location relative to nodeOrigin
	node->translate.y	= 0.0f;
	node->translate.z	= 0.0f;

	node->nodeOrigin.x	= 0.0f;				//origin relative to parent origin
	node->nodeOrigin.y	= 0.0f;
	node->nodeOrigin.z	= 0.0f;
	
	node->rotateRate.x	= 0.0f;
	node->rotateRate.y	= 0.0f;
	node->rotateRate.z	= 0.0f;

	node->rotateRad.x	= 0.0f;
	node->rotateRad.y	= 0.0f;
	node->rotateRad.z	= 0.0f;
	
	node->scaleRate.x	= 0.0f;				//scale rate 
	node->scaleRate.y	= 0.0f;
	node->scaleRate.z	= 0.0f;

	node->translateRate.x = 0.0f;			//pans the location of the node, pronunciar en Espanol, lol
	node->translateRate.y = 0.0f;
	node->translateRate.z = 0.0f;

	node->translateVect.x = 0.0f;			//pans the location of the node, pronunciar en Espanol, lol
	node->translateVect.y = 0.0f;
	node->translateVect.z = 0.0f;
	
	node->shader	= 1;				// texture ID and render type
	node->geometry		= 0;//kSurfaceNull;		//FFT layers, mesh, sphere...

	node->lineWidth		= 1.0f;				//draws lines, zero for off
	node->pointSize		= 0.0f;				//draws dots, zero for off
	node->ratio			= kNPdefaultRatio;	//torus inner radius

	node->colorIndex	= 0;
	
	node->color.r		= 50;				//default color assigned to new data
	node->color.g		= 101;
	node->color.b		= 101;
	node->color.a		= 255;				//fully opaque

	node->colorFade		= false;
	node->textureID		= 0;				//zero for no texture

	node->hide			= false;			//hide the node, data stays active
	node->freeze		= false;			//freezes data and physics updates
	node->center		= false;			//centers to nodeOrigin

	node->autoZoom.x	= false;			//scales node to fit screen
	node->autoZoom.y	= false;
	node->autoZoom.z	= false;

	node->scroll		= false;			//scroll right to left
	
	node->triggerOnHi.x	= false;			//triggers are multipurpose
	node->triggerOnHi.y	= false;
	node->triggerOnHi.z	= false;

	node->triggerOnLow.x = false;
	node->triggerOnLow.y = false;
	node->triggerOnLow.z = true;			//prevent from going below ground
	
	node->triggerLevelHi.x = 0.0f;
	node->triggerLevelHi.y = 0.0f;
	node->triggerLevelHi.z = 0.0f;

	node->triggerLevelLow.x = 0.0f;
	node->triggerLevelLow.y = 0.0f;
	node->triggerLevelLow.z = 0.0f;

	node->proximity.x	= 88;			//used for collision engine 
	node->proximity.y	= false;
	node->proximity.z	= false;
	
	node->proximityRadius.x = 0.0f;
	node->proximityRadius.y = 0.0f;
	node->proximityRadius.z = 0.0f;

	node->segments.x		= 11;
	node->segments.y		= 22;
	node->segments.z		= 33;

	node->topo				= 44;

	node->formatID			= 55;			//maps db fields and translates
	node->tableID			= 66;			//stores the source table location
	node->recordID			= 0;			//source db record
	
	node->size = 420;//sizeof(NPnode);		//causes problem with CSV loading, debug zz

	// below should possbily be removed, not needed in CSV or DB file
	node->screenXY.x		= 0;							//MB-LABEL
	node->screenXY.y		= 0;							//MB-LABEL
}


// cameras can be GL scene cameras or real-world cameras
// video is attached as a daughter of the recNode
//-----------------------------------------------------------------------------
void npInitNodeCamera (NPnodePtr node)
{
	NPcameraPtr data = NULL;
	
	//allocate memory for camera data struct if does not already exist
	if (node->data == NULL)
	{
		data = (NPcameraPtr) malloc (sizeof(NPcamera));
		if (data == NULL)
		{
			printf ("err 4490 - malloc failed, out of memory \n");
			return;
		}
		node->data  = (void*) data;
	}

	//main node params
	node->type			 = kNodeCamera;			//defines the node type
	node->triggerOnLow.z = false;				//allows cam below surface

	//init cam data
	data->id			= node->id;
	data->type			= kNodeCamera;

	data->format		= kVideoFormat720p;
	data->interlaced	= kVideoFieldNull;		//progressive scan, no fields
	data->stereo		= kVideoStereoNull;		//not stereoscopic
	data->aspectRatio	= 1.77778f;				//1.77778 is approx 16:9
	data->colorSpace	= kColorSpaceRGB_8;		//8bit RGB is what GPU's output
	data->fps			= 59.97f;
	data->width			= 1280;
	data->height		= 720;

	data->fov			= 35.0f;				//35mm, 70mm...
	data->clipNear		= 1.0f;					//clipping planes
	data->clipFar		= 1000.0f;

	data->aperture		= 1.0f;					//F stop
	data->exposure		= 0.01f;				//in seconds
	data->sensorType	= kCameraSensorNull;	//3CCD, Debayer pattern...

	data->size = sizeof(NPcamera);
}


//-----------------------------------------------------------------------------
void InitNodeVideo (NPnodePtr node)
{
	NPvideoPtr data = (NPvideoPtr) malloc (sizeof(NPvideo));
	if (data == NULL)
	{
		printf ("\n 4214 error malloc failed cannot write file \n");
		return;
	}

	node->data  = (void*) data;

	node->type			= kNodeVideo;

	data->id			= node->id;
	data->type			= kNodeVideo;

	data->format		= kVideoFormat720p;
	data->interlaced	= kVideoFieldNull;		//progressive scan, no fields
	data->stereo		= kVideoStereoNull;		//not stereoscopic
	data->aspectRatio	= 1.77778f;				//1.77778 is approx 16:9
	data->colorSpace	= kColorSpaceRGB_8;		//8bit RGB is what GPU's output
	data->fps			= 59.97f;
	data->width			= 1280;
	data->height		= 720;

	data->brightness	= 1.0f;				// 1.0 center, range 0.0 to 100.0
	data->contrast		= 1.0f;				// 1.0 center, range 0.0 to 100.0
	data->hue			= 0.0f;				// 0.0 center, range -1.0 to +1.0
	data->saturation	= 1.0f;				// 1.0 center, 0.0 is BW, max 100.0
	data->blackLevel	= 0.0f;				// 0.0 and 7.5 IRE typical

	data->bias.r		= 0.0f;				//brightness shift, 0.0 no change
	data->bias.g		= 0.0f;
	data->bias.b		= 0.0f;
	data->bias.a		= 0.0f;
	data->gain.r		= 1.0f;				//color multiplier, 1.0 no change
	data->gain.g		= 1.0f;
	data->gain.b		= 1.0f;
	data->gain.a		= 1.0f;

	data->size = sizeof(NPvideo);
}


//------------------------------------------------------------------------------
void InitNodeSurface (NPnodePtr node)
{
	NPsurfacePtr data = (NPsurfacePtr) malloc (sizeof(NPsurface));
	if (data == NULL)
	{
		printf ("\n 4215 error malloc failed cannot write file \n");
		return;
	}

	node->data  = (void*) data;

	node->type			= kNodeSurface;

	data->id			= node->id;
	data->type			= kNodeSurface;

	data->vertexArray	= NULL;
	data->normalArray	= NULL;
	data->colorArray	= NULL;
	data->vertexCount	= 0;
	data->normalCount	= 0;
	data->arrayIndex	= 0;

	data->size = sizeof(NPsurface);
}


//-----------------------------------------------------------------------------
void InitNodePoints (NPnodePtr node)
{
	NPpointsPtr data = (NPpointsPtr) malloc (sizeof(NPpoints));
	if (data == NULL)
	{
		printf ("\n 4216 error malloc failed cannot write file \n");
		return;
	}

	node->data  = (void*) data;
	
	node->type			= kNodePoints;

	data->id			= node->id;
	data->type			= kNodePoints;

	data->vertexArray	= NULL;				//the array of data points
	data->colorArray	= NULL;				//each point gets color assigned
	data->vertexCount	= 0;				//number of points in the node
	data->arrayIndex	= 0;				//current vertex index in the array

	data->size = sizeof(NPpoints);
}


//-----------------------------------------------------------------------------
void InitNodePin (NPnodePtr node)
{
	NPpinPtr data = (NPpinPtr) malloc (sizeof(NPpin));
	if (data == NULL)
	{
		printf ("\n 4217 error malloc failed cannot write file \n");
		return;
	}

	node->data  = (void*) data;


	node->type			= kNodePin;
	node->geometry		= kNPprimitivePin;		//ice-cream cone shape

	data->id			= node->id;

	data->innerRadius	= 0.1f;			//inner radius of the toroid
	data->outerRadius	= 1.0f;			//outer radius of the toroid

	// 4 * outerRadius / innerRadius -> int
	data->radiusRatioIndex = 40;			//sorts pre-loaded GPU toriods
	data->screenSizeIndex = 8;				//same as above for detail level

	data->slices		= 7;				//calculated from screenSizeIndex
	data->stacks		= 16;

	data->translateTex.x = 0.0f;			//position offset of texture map
	data->translateTex.y = 0.0f;
	data->translateTex.z = 0.0f;

	data->scaleTex.x	= 1.0f;
	data->scaleTex.y	= 1.0f;
	data->scaleTex.z	= 1.0f;

	data->rotateTex.x	= 0.0f;
	data->rotateTex.y	= 0.0f;
	data->rotateTex.z	= 0.0f;
	data->rotateTex.angle = 0.0f;			//orientation of texture map

	data->size = sizeof(NPpin);
}


//------------------------------------------------------------------------------
void InitNodeGrid (NPnodePtr node)
{
	NPgridPtr gridData = (NPgridPtr) malloc (sizeof(NPgrid));

	if (gridData == NULL)
	{
		printf ("\n 4218 error malloc failed cannot write file \n");
		return;
	}

	node->data = gridData;

	node->type			= kNodeGrid;

	node->colorIndex	= 3;
	node->color.r		= 0;
	node->color.g		= 0;
	node->color.b		= 255;
	node->color.a		= 255;

	gridData->id			= node->id;
	gridData->type			= kNodeGrid;

	node->segments.x = kNPgridSegments;
	node->segments.y = kNPgridSegments;
	node->segments.z = 0;

	gridData->spacing.x		= kNPgridSpacing;
	gridData->spacing.y		= kNPgridSpacing;
	gridData->spacing.z		= kNPgridSpacing;

	gridData->overlay		= false;

	gridData->size			= sizeof(NPgrid);
}

//------------------------------------------------------------------------------
int npNewNodeID()				//improve this to re-use nodeID upon delete, debug zz
{
	static int nodeID = 0;

	nodeID++;
												// 2^24 max for picking algorithm
	if (nodeID >= kNPnodeMax)
		nodeID = kNPnodeMax - 1;

	return nodeID;
}

	
//return the selected camera based on the root camera childIndex
//------------------------------------------------------------------------------
NPnodePtr cameraSelected(void)
{
	pData data = (pData) npGetDataRef();
	NPnodePtr camNode = data->map.node[kNPnodeRootCamera];

	camNode = camNode->child[camNode->childIndex];
	
	return camNode;
}

