//------------------------------------------------------------------------------
//
//  npgldraw.c
//
// ANTZ - 3D data viewer with realtime real-world interaction based on NPE.
//
// ANTZ is hosted at http://temporalzone.com and NPE at http://neuralphysics.org
//
// Written in 2010-2011 by Shane Saxon - contact@saxondigital.net
//
// Additional code contributed by: Mark Buchanan and Lucas Erickson.
//
// To the extent possible under law, the author(s) have dedicated all copyright 
// and related and neighboring rights to this software to the public domain
// worldwide. This software is distributed without any warranty.
//
// Released under the CC0 license, which is GPL compatible.
//
// You should have received a copy of the CC0 Public Domain Dedication along
// with this software (license file named COPYING.txt). If not, see
// http://creativecommons.org/publicdomain/zero/1.0/
//
//------------------------------------------------------------------------------

#include "npgldraw.h"

#include "npgeometry.h"

#include "../../npdata.h"
#include "../../npctrl.h"
#include "../npglut.h"


void DrawDefault (NPnodePtr node);
void DrawCamera (NPnodePtr node);
void DrawSurface (NPnodePtr node);
void DrawPoints (NPnodePtr node);
void DrawPin (int selected, NPnodePtr node, void* dataRef);
void DrawPinChild (NPnodePtr node, void* dataRef);
void DrawVideo (NPnodePtr node);
void DrawGrid (NPnodePtr node, void* dataRef);
void npDrawSort(void* dataRef);


//------------------------------------------------------------------------------
void npInitGLDraw(void* dataRef)
{
	pData data = (pData) dataRef;

//	InitTorus (dataRef);

	data->io.gl.pickPass = false;			//zz global init
}


//------------------------------------------------------------------------------
void DrawDefault (NPnodePtr node)
{
	return;
}


// draw an icon at the location of the camera in the scene
//------------------------------------------------------------------------------
void DrawCamera (NPnodePtr node)
{
	// draw icon of camera in the scene, but not the current viewport camera
	return;
}


//------------------------------------------------------------------------------
void DrawSurface (NPnodePtr node)
{
	// draw ground
	glPushMatrix();

	// position the node
	glTranslatef (node->translate.x, node->translate.y, node->translate.z);
	glRotatef (node->rotate.angle, node->rotate.x, node->rotate.y, node->rotate.z);
	glScalef (node->scale.x, node->scale.y, node->scale.z);

	glColor4ub (node->color.r, node->color.g, node->color.b, node->color.a);

	glBegin(GL_QUADS);
		glVertex3f (-100.0f, -100.0f, 0.0f);
		glVertex3f (-100.0f,  100.0f, 0.0f);
		glVertex3f ( 100.0f,  100.0f, 0.0f);
		glVertex3f ( 100.0f, -100.0f, 0.0f);
	glEnd();

	glPopMatrix();
}


//------------------------------------------------------------------------------
void DrawPoints (NPnodePtr node)
{
	return;
}


//MB//

// converts the ID value to an RGB value for picking objects by color
//------------------------------------------------------------------------------
void npIDtoRGB( int id, int* r, int* g, int* b )
{
	int remain = id;
	
	*r = remain / (256*256);	remain = remain - *r * (256*256);
	*g = remain / 256;			remain = remain - *g * 256;
	*b = remain;
}

int npRGBtoID( int r, int g, int b )
{
	int id;
	id = (256*256)*r + 256*g + b;

	return id;
}


void PickPass( int picking, void* dataRef )
{
	pData data = (pData) dataRef;

	data->io.gl.pickPass = picking;
}

int npGetPickPass(void* dataRef)
{
	pData data = (pData) dataRef;

	return data->io.gl.pickPass;
}

//MB-End//

//------------------------------------------------------------------------------
void DrawPin (int currentNode, NPnodePtr node, void* dataRef)
{
	int i = 0;
	int branchLevel = 0;

	pData data = (pData) dataRef;

	NPnodePtr nodeChild = data->map.activeNode;

	//MB//	
	int idRed, idGrn, idBlu;
	int remain = 0;
	idRed = idGrn = idBlu = 0;
	//-

	branchLevel = nodeChild->branchLevel;


	//MB//
	//if pickPass render scene with node ID encoded as color
	if( data->io.gl.pickPass )
	{
		if( data->io.gl.pickID == node->id )	//move ouside draw process, debug zz
		{										//then no need to draw scene twice, zz
			//select the proper root
			currentNode = true;

			data->map.nodeRootIndex =  npGetRootIndex (node, dataRef);
			data->map.selectedIndex = data->map.nodeRootIndex;
		}
		
		//Convert nodeID to RGB triplet
		npIDtoRGB( node->id, &idRed, &idGrn, &idBlu );
		glColor4ub( idRed, idGrn, idBlu, 255 );	//ID of Node is encoded on the Pixel! In RGB Triplet   (Range: 0-  16.7 Million)
	}
	else
		glColor4ub (node->color.r, node->color.g, node->color.b, node->color.a);


	glPushMatrix();

	// position the node
	glTranslatef (node->translate.x, node->translate.y, node->translate.z);
	glRotatef (node->rotate.angle, node->rotate.x, node->rotate.y, node->rotate.z);
	glScalef (node->scale.x, node->scale.y, node->scale.z);

	glLineWidth(node->lineWidth);

	if (node->geometry == kNPprimitivePin)
	{
		npGLSurface (node->textureID, node, data);	
		glTranslatef (0.0f, 0.0f, 5.0f);			//should this be here, debug, zz
	}
	else
	{
		glTranslatef (0.0f, 0.0f, 5.0f);
		npGLSurface (node->textureID, node, data);
	}

	glLineWidth(1.0f);	//zz

	if (node->selected || currentNode)
	{
		glPushMatrix();

		if (currentNode)
		{
			// slide down a little for scaling up to surround the pin
			glTranslatef (0.0f, 0.0f, -5.10f);
			glScalef (1.15f, 1.15f, 1.15f);


			if ( node->selected )
			{
				if( !data->io.gl.pickPass )
					glColor4f ( 0.8f, 0.8f, 0.0f, 1.0f);	// draw selected yellow
				npGLPrimitive (kNPprimitiveWirePin, 0.0f);
			}
			else if (branchLevel == 0)
			{
				if( !data->io.gl.pickPass )
					glColor4f ( 1.0f, 0.0f, 0.0f, 1.0f);	// current node red
				npGLPrimitive (kNPprimitiveWirePin, 0.0f);
			}

			if (branchLevel > 0 && !data->io.gl.pickPass)
			{
				if (nodeChild->selected)
					glColor4f ( 0.8f, 0.8f, 0.0f, 1.0f);	// draw selected yellow
				else
					glColor4f ( 1.0f, 0.0f, 0.0f, 1.0f);	// current node red
			}
			glTranslatef (0.0f, 0.0f, 5.0f);
			glScalef (1.0f, 1.0f, 0.66f);
			npGLPrimitive (kNPprimitiveWirePin, 0.0f);		// 2nd one will be above
		}
		else if (node->selected)
		{
			if( !data->io.gl.pickPass )
				glColor4f ( 0.8f, 0.8f, 0.0f, 1.0f);		// draw selected as yellow

			glTranslatef (0.0f, 0.0f, -5.10f);
			glScalef (1.05f, 1.05f, 1.05f);
			npGLPrimitive (kNPprimitiveWirePin, 0.0f);
		}

		glPopMatrix();
	}

	//should call DrawNode, that calls draw pin, move this to drawnodes...   debug, zz
	// recursively
	for (i=0; i < node->childCount; i++)
	{
		glPushMatrix();
			DrawPinChild (node->child[i], dataRef);
		glPopMatrix();
	}

	glPopMatrix();
}

// first draw child here, then break it up, draw nodes should iterate down, debug, zz
// the tree branches and take care of selected glows, etc...
// this routine should be generic regardless of branch level
//------------------------------------------------------------------------------
void DrawPinChild (NPnodePtr node, void* dataRef)
{
	int i = 0;
	int idRed = 0, idGrn = 0, idBlu = 0;

	pData data = (pData) dataRef;

	NPpinPtr nodeData = (NPpinPtr) node->data;


	//Main Color for Rendering Pin
	if( data->io.gl.pickPass )
	{
		if (data->io.gl.pickID == node->id )	//move ouside draw process, debug zz
		{
			data->map.nodeRootIndex = npGetRootIndex (node, dataRef);
			data->map.selectedIndex = data->map.nodeRootIndex;
		}

		npIDtoRGB( node->id, &idRed, &idGrn, &idBlu );
		glColor4ub( idRed, idGrn, idBlu, 255 );	//ID of Node is encoded on the Pixel! In RGB Triplet   (Range: 0-  16.7 Million)
	}
	else
		glColor4ub (node->color.r, node->color.g, node->color.b, node->color.a);


	if (node->branchLevel > 1)
	{
		//position
		glRotatef (90.0f, 1.0f, 0.0f, 0.0f);
		glRotatef (node->rotate.angle, 0.0f, 1.0f, 0.0f);		//x	debug zz
			
		glTranslatef ( (node->translate.x * node->scale.x) + 1.5f,
						node->translate.y * node->scale.y,
						0.0f );								// z not translated

		//orientation
		glRotatef (node->rotateRad.y * 57.29578f, 0.0f, 0.0f, 1.0f);		//y, debug zz
		glRotatef (node->rotateRad.z * 57.29578f, 1.0f, 0.0f, 0.0f);		//y, debug zz

		glScalef ( 0.5f * node->scale.x, 0.5f * node->scale.y, 0.5f * node->scale.z);
	}
	else
	{
		glTranslatef (node->translate.x * node->scale.x,
						node->translate.y * node->scale.y,
						node->translate.z );				// z is not scaled

		glRotatef (node->rotateRad.y * 57.29578f, 0.0f, 0.0f, 1.0f);		//y, debug zz
		glScalef ( node->scale.x, node->scale.y, node->scale.z);	//root child
	}


	// draw the node unless hidden
	if (!node->hide)
	{
		glLineWidth(node->lineWidth);
		npGLSurface (node->textureID, node, data);
		glLineWidth(1.0f);
	}

	// draw a selection status wireframes slightly scaled up on top...
	if (node->selected)
	{	
		glPushMatrix();
			// draw nodeID color on pickPass to allow picking on the wireframe
			if( !data->io.gl.pickPass )
				glColor4f (0.8f, 0.8f, 0.0f, 1.0f);			//draw yellow
			glScalef (1.03f, 1.03f, 1.03f);
			npGLPrimitive (kNPglutWireTorus, node->ratio);
		glPopMatrix();
	}

	// draw a wireframe slightly scaled up on top...
	if (node == data->map.selectedNode)
	{	
		glPushMatrix();
			// draw nodeID color on pickPass to allow picking on the wireframe
			if( !data->io.gl.pickPass )
				glColor4f (1.0f, 0.0f, 0.0f, 1.0f);			//draw red
			glScalef (1.04f, 1.04f, 1.04f);
			npGLPrimitive (kNPglutWireTorus, node->ratio);
		glPopMatrix();
	}

	//draw points slightly scaled up on top...
	if (node->pointSize > 0.0f)
	{
		//draw points slightly scaled up on top...		//add feature, debug zz
	}

	
	// draw the child nodes recursively
	for (i=0; i < node->childCount; i++)
	{
		glPushMatrix();
			DrawPinChild (node->child[i], data);
		glPopMatrix();	
	}
}


//------------------------------------------------------------------------------
void DrawVideo (NPnodePtr node)
{
	return;
}


//------------------------------------------------------------------------------
void DrawGrid (NPnodePtr node, void* dataRef)
{
	int			i = 0;
	int			j = 0;
	int			k = 0;
	int			idRed = 0, idGrn = 0, idBlu = 0;
	GLfloat		x = 0.0f;
	GLfloat		y = 0.0f;
	GLfloat		z = 0.0f;
	NPfloatXYZ	length;

	NPnodePtr camNode = cameraSelected();
	NPgridPtr grid = (NPgridPtr) node->data;
	pData data = (pData) dataRef;

	//MB-TEXTURE
	static GLfloat sgenparams[] = { +1.0,  0.0,  0.0,  0.0 };
	static GLfloat tgenparams[] = {  0.0, +1.0,  0.0,  0.0 };
	//MB-END

	//if pickPass and pickMode not grid then don't draw grid
	if (data->io.gl.pickPass && data->io.gl.pickMode != kNPpickModeGrid)
		return;

	glDisable (GL_LIGHTING);	//draw 100% ambient white
	
	glPushMatrix();				//restore after calling child nodes

	//MB-TEXTURE
	glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL ); //GL_MODULATE
	glTexGeni( GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR ); //GL_SPHERE_MAP
	glTexGeni( GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR );
	glTexGenfv( GL_S, GL_OBJECT_PLANE, sgenparams );
	glTexGenfv( GL_T, GL_OBJECT_PLANE, tgenparams );
	glDisable (GL_TEXTURE_GEN_S);
	glDisable (GL_TEXTURE_GEN_T);
	//MB-END

	//calculate grid size based on segment count and spacing
	length.x = grid->spacing.x * node->segments.x;
	length.y = grid->spacing.y * node->segments.y;
	length.z = grid->spacing.z * node->segments.z;

	// position node
	glTranslatef (node->translate.x, node->translate.y, node->translate.z);
	glRotatef (node->rotate.angle, node->rotate.x, node->rotate.y, node->rotate.z);
	glScalef (node->scale.x, node->scale.y, node->scale.z);

	//center based on length
	glTranslatef (length.x * -0.5f, length.y * -0.5f, 0.0f);

	// if pickPass then set color by id
	if (data->io.gl.pickPass)
	{
		npIDtoRGB (node->id, &idRed, &idGrn, &idBlu);
		glColor4b (idRed, idGrn, idBlu, 255);		//draw quad color mapped as id
	}

	//Draw Texture Quad
	if (node->textureID)	//allow for click on background tex's for mouse flying
	{
		//if not pickPass then enable and bind texture, set color to white
		if (!data->io.gl.pickPass)
		{
			glEnable (GL_TEXTURE_2D);
			glBindTexture (GL_TEXTURE_2D, node->textureID);
			glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
		}

		//draw the quad
		glBegin (GL_QUADS);
			glTexCoord2d(0.0,0.0); glVertex3d(0.0f,  0.0f, 0.0f);	 // center quad
			glTexCoord2d(1.0,0.0); glVertex3d(length.x, 0.0f, 0.0f);
			glTexCoord2d(1.0,1.0); glVertex3d(length.x, length.y, 0.0f);
			glTexCoord2d(0.0,1.0); glVertex3d(0.0f,  length.y, 0.0f);
		glEnd();

		glDisable ( GL_TEXTURE_2D );
	}

	// if not pickPass then set the node color
	if (!data->io.gl.pickPass)
		glColor4ub (node->color.r, node->color.g, node->color.b, node->color.a);
	
	//draw grid lines on top of texture map
	glDisable (GL_DEPTH_TEST);
	glLineWidth (node->lineWidth);

	//draw grid lines
	glBegin (GL_LINES);
		for (i=0; i <= node->segments.z; i++)
		{
			for (j=0; j <= node->segments.x; j++)
			{
				glVertex3f (x, 0.0f, z);
				glVertex3f (x, length.y, z);
				x += grid->spacing.x;
			}
			for (j=0; j <= node->segments.y; j++)
			{
				glVertex3f (0.0f, y, z);
				glVertex3f (length.x, y, z);
				y += grid->spacing.y;
			}

			x = 0.0f;
			y = 0.0f;
			z += grid->spacing.z;
		}
	glEnd();

	glLineWidth (1.0f);
	glEnable (GL_DEPTH_TEST);

	//undo center based on length
	glTranslatef (length.x * 0.5f, length.y * 0.5f, 0.0f);

	//recursively traverse and draw children
	for (i=0; i < node->childCount; i++)
		DrawGrid(node->child[i], data);

	glPopMatrix();				//restore after traversing child nodes

	//restore lighting
	if (data->io.gl.pickPass == false)
		glEnable (GL_LIGHTING);
}


// draw 2D overlays including grids, text, messages, highlight objects, etc...
//------------------------------------------------------------------------------
void npDrawHUD (void* dataRef)
{
	return;
}


// sets the node RGB color based on index, preserves existing alpha value
//------------------------------------------------------------------------------
void SetIndexColor (NPubyteRGBA *color, int *colorIndex)
{	
	int index = *colorIndex;
									//kNPpaletteSize stored in nptypes.h
	static const GLubyte colorPallete[kNPpaletteSize][3] = {
											{50,  101, 101},
											{0,   255, 0},
											{255,   0, 0},
											{0,     0, 255},
											{255, 255, 0},
											{152,   0, 255},
											{255, 168, 0},
											{0,   255, 255},
											{255,   0, 255},
											{0,   153, 0},
											{173, 255, 202},
											{255, 180, 255},
											{0,   152, 255},
											{185, 255, 0},
											{152,   0, 0},
											{127, 127, 127},
											{127, 127, 255},
											{197,  82, 0},
											{0,     0, 0},
											{255, 255, 255}
											};

	//re-maps out of bounds colorIndex
	if (index >= kNPpaletteSize)
	{
		while (index >= kNPpaletteSize)
			index -= kNPpaletteSize;		//subs pallete sized increments	
		*colorIndex = index;
	}

	if (index < 0)
	{
		while (index < 0)
			index += kNPpaletteSize;	
		*colorIndex = index;			//adds pallete sized increments				
	}

	//now set the color
	color->r = colorPallete[index][0];
	color->g = colorPallete[index][1];
	color->b = colorPallete[index][2];
}




//MB-Transp

//------------------------------------------------------------------------------
//Distance between two points
float npDistance( NPfloatXYZ vecA, NPfloatXYZ vecB )
{
	float x = vecA.x - vecB.x;
	float y = vecA.y - vecB.y;
	float z = vecA.z - vecB.z;
	
	float len = x*x + y*y + z*z;
	len = (float)sqrt( len );
	return len;
}

//------------------------------------------------------------------------------
//QSort Compare function
int comparePinsDistanceFromCamera (const void* a, const void* b)
{
	NPnodePtr* pinAptr = (NPnodePtr*)a;
	NPnodePtr* pinBptr = (NPnodePtr*)b;
	NPnodePtr pinA = *pinAptr;
	NPnodePtr pinB = *pinBptr;
	if( pinA == NULL && pinB == NULL )	//Null element in pinA & pinB: Nothing or Grid or Camera
		return 0;						//These are "equal". Both should be at the beginning of the list.
	if( pinA == NULL )					//Null element in pinA: Nothing or Grid or Camera
		return -1;						//pinA is at beginning of list... therefore pinA is    "less than" pinB. Return -1
	if( pinB == NULL )					//Null element in pinB: Nothing or Grid or Camera
		return +1;						//pinB is at beginning of list... therefore pinA is "greater than" pinB. Return +1
	
	if( pinA->distFromCamera   >  pinB->distFromCamera )	//pinA is furter from camera than pinB.
		return -1;											//Therefore should be at the beginning of the list. ie "less than"... Return -1
	if( pinA->distFromCamera   <  pinB->distFromCamera )	//pinA is closer from camera than pinB.
		return +1;											//Therefore should be at the end of the list.    ie "greater than"... Return +1
	if( pinA->distFromCamera  ==  pinB->distFromCamera )	//pinA & pinB are exactly the same distance from the camera. (impossible-ish)
		return 0;											//Return 0;

	//return ( *(int*)a - *(int*)b );
	return -1;
}


//------------------------------------------------------------------------------
void npDrawSort(void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;

	NPnodePtr node = NULL;
	NPnodePtr camNode = cameraSelected();

	int nodeRootCount = data->map.nodeRootCount;

	NPfloatXYZ camVec = camNode->translate;

	//update root node distance from camera
	for( i = 0; i < nodeRootCount; i++ )
	{
		if( i < kNPnodeRootPin )	//Nothing, Camera, Grid
		{
			data->map.sort[i] = NULL;
			continue;
		}
		
		node = data->map.node[i];
		node->distFromCamera = npDistance( camVec, node->translate );

		data->map.sort[i] = node;
	}

	//create z-sorted list, back to front
	//QuickSort based on Distance From Camera, back to front z-sorted
	qsort( data->map.sort, nodeRootCount, sizeof(NPnodePtr),
			comparePinsDistanceFromCamera );

}

//------------------------------------------------------------------------------
void npDrawNodes (void* dataRef)
{
	int i = 0;
	int selected = false;

	NPnodePtr node = NULL;
	pData data = (pData) dataRef;

	int count = data->map.nodeRootCount;

	//sort nodes, back to front z-sort for transparency
	npDrawSort(data);

	//draw all nodes
	for( i=0; i < count; i++ )
	{
		if( i < kNPnodeRootPin )
			node = data->map.node[i];	//first draw camera and grids
		else
			node = data->map.sort[i];	//draw in sorted order
		
		if (node == data->map.node[data->map.nodeRootIndex]) //use seletedIndex? debug zz
			selected = true;
		else
			selected = false;

		if (!node->hide)				//if not hidden draw node
			switch (node->type)
		{
			case kNodeDefault	: DrawDefault(node); break;
			case kNodeCamera	: DrawCamera (node); break;
			case kNodeSurface	: DrawSurface(node); break;
			case kNodePoints	: DrawPoints (node); break;
			case kNodePin		: DrawPin (selected, node, data); break;
			case kNodeVideo		: DrawVideo  (node); break;
			case kNodeGrid		: DrawGrid   (node, data); break;
		}
	}
}

//MB-LABEL
void DrawSimpleTorus(void)
{
	npGLSurface (kNPglutSolidTorus, 0, npGetDataRef());
}

